@@ -6,10 +6,13 @@ import Common as X
 import Types as X
 import Key as X
 import Types.UUID as X
-import Annex as X (gitRepo, inRepo, fromRepo, calcRepo)
+import Annex as X (gitRepo, inRepo, fromRepo, calcRepo, calcRepo')
 import Annex.Locations as X
 import Annex.Debug as X (fastDebug, debug)
 import Messages as X
+import Git.Quote as X
+import Types.RepoSize as X
+import Git.Types as X (RemoteName)
 #ifndef mingw32_HOST_OS
-import System.Posix.IO as X hiding (createPipe)
+import System.Posix.IO as X hiding (createPipe, append)
 #endif
