@@ -1,11 +1,12 @@
 {- git-annex import from remotes
  -
- - Copyright 2019-2024 Joey Hess <id@joeyh.name>
+ - Copyright 2019-2025 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
 
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE CPP #-}
 
 module Annex.Import (
 	ImportTreeConfig(..),
@@ -20,6 +21,7 @@ module Annex.Import (
 	importKeys,
 	makeImportMatcher,
 	getImportableContents,
+	PostExportLogUpdate,
 ) where
 
 import Annex.Common
@@ -37,12 +39,12 @@ import qualified Annex
 import Annex.Link
 import Annex.LockFile
 import Annex.Content
-import Annex.Export
 import Annex.RemoteTrackingBranch
 import Annex.HashObject
 import Annex.Transfer
 import Annex.CheckIgnore
 import Annex.CatFile
+import Annex.GitShaKey
 import Annex.VectorClock
 import Annex.SpecialRemote.Config
 import Command
@@ -63,14 +65,16 @@ import qualified Utility.Matcher
 import qualified Database.Export as Export
 import qualified Database.ContentIdentifier as CIDDb
 import qualified Logs.ContentIdentifier as CIDLog
+import qualified Utility.OsString as OS
 import Backend.Utilities
 
 import Control.Concurrent.STM
 import qualified Data.Map.Strict as M
 import qualified Data.Set as S
-import qualified System.FilePath.Posix.ByteString as Posix
-import qualified System.FilePath.ByteString as P
 import qualified Data.ByteArray.Encoding as BA
+#ifdef mingw32_HOST_OS
+import qualified System.FilePath.Posix as Posix
+#endif
 
 {- Configures how to build an import tree. -}
 data ImportTreeConfig
@@ -109,8 +113,9 @@ buildImportCommit
 	-> ImportCommitConfig
 	-> AddUnlockedMatcher
 	-> Imported
+	-> PostExportLogUpdate
 	-> Annex (Maybe Ref)
-buildImportCommit remote importtreeconfig importcommitconfig addunlockedmatcher imported =
+buildImportCommit remote importtreeconfig importcommitconfig addunlockedmatcher imported postexportlogupdate =
 	case importCommitTracking importcommitconfig of
 		Nothing -> go Nothing
 		Just trackingcommit -> inRepo (Git.Ref.tree trackingcommit) >>= \case
@@ -118,12 +123,14 @@ buildImportCommit remote importtreeconfig importcommitconfig addunlockedmatcher
 			Just _ -> go (Just trackingcommit)
   where
 	go trackingcommit = do
-		(importedtree, updatestate) <- recordImportTree remote importtreeconfig (Just addunlockedmatcher) imported
+		(importedtree, updatestate) <- recordImportTree remote importtreeconfig (Just addunlockedmatcher) imported postexportlogupdate
 		buildImportCommit' remote importcommitconfig trackingcommit importedtree >>= \case
 			Just finalcommit -> do
 				updatestate
 				return (Just finalcommit)
-			Nothing -> return Nothing
+			Nothing -> do
+				postExportLogUpdate postexportlogupdate
+				return Nothing
 
 {- Builds a tree for an import from a special remote.
  -
@@ -135,8 +142,9 @@ recordImportTree
 	-> ImportTreeConfig
 	-> Maybe AddUnlockedMatcher
 	-> Imported
+	-> PostExportLogUpdate
 	-> Annex (History Sha, Annex ())
-recordImportTree remote importtreeconfig addunlockedmatcher imported = do
+recordImportTree remote importtreeconfig addunlockedmatcher imported postexportlogupdate = do
 	importedtree@(History finaltree _) <- buildImportTrees basetree subdir addunlockedmatcher imported
 	return (importedtree, updatestate finaltree)
   where
@@ -154,7 +162,7 @@ recordImportTree remote importtreeconfig addunlockedmatcher imported = do
 				let subtreeref = Ref $
 					fromRef' finaltree
 						<> ":"
-						<> getTopFilePath dir
+						<> fromOsPath (getTopFilePath dir)
 				in fromMaybe emptyTree
 					<$> inRepo (Git.Ref.tree subtreeref)
 		updateexportdb importedtree
@@ -177,6 +185,7 @@ recordImportTree remote importtreeconfig addunlockedmatcher imported = do
 			{ oldTreeish = exportedTreeishes oldexport
 			, newTreeish = importedtree
 			}
+		postExportLogUpdate postexportlogupdate
 		return oldexport
 
 	-- importKeys takes care of updating the location log
@@ -349,11 +358,11 @@ mkImportTreeItem maddunlockedmatcher msubdir loc v = case v of
 	lf = fromImportLocation loc
 	treepath = asTopFilePath lf
 	topf = asTopFilePath $
-		maybe lf (\sd -> getTopFilePath sd P.</> lf) msubdir
+		maybe lf (\sd -> getTopFilePath sd </> lf) msubdir
 	mklink k = do
 		relf <- fromRepo $ fromTopFilePath topf
 		symlink <- calcRepo $ gitAnnexLink relf k
-		linksha <- hashSymlink symlink
+		linksha <- hashSymlink (fromOsPath symlink)
 		return $ TreeItem treepath (fromTreeItemType TreeSymlink) linksha
 	mkpointer k = TreeItem treepath (fromTreeItemType TreeFile)
 		<$> hashPointerFile k
@@ -429,7 +438,12 @@ buildImportTreesGeneric converttree basetree msubdir importable@(ImportableConte
 		-- Full directory prefix where the sub tree is located.
 		let fullprefix = asTopFilePath $ case msubdir of
 			Nothing -> subdir
-			Just d -> getTopFilePath d Posix.</> subdir
+			Just d ->
+#ifdef mingw32_HOST_OS
+				toOsPath $ fromOsPath (getTopFilePath d) Posix.</> fromOsPath subdir
+#else
+				getTopFilePath d </> subdir
+#endif
 		Tree ts <- converttree (Just fullprefix) $
 			map (\(p, i) -> (mkImportLocation p, i))
 				(importableContentsSubTree c)
@@ -490,11 +504,26 @@ canImportKeys remote importcontent =
   where
 	ia = Remote.importActions remote
 
--- Result of an import. ImportUnfinished indicates that some file failed to
--- be imported. Running again should resume where it left off.
+-- Result of an import. 
 data ImportResult t
-	= ImportFinished t
+	= ImportFinished PostExportLogUpdate t
 	| ImportUnfinished
+	-- ^ ImportUnfinished indicates that some file failed to
+	-- be imported. Running again should resume where it left off.
+
+-- An action to run after the export log has been updated to reflect an
+-- import.
+newtype PostExportLogUpdate = PostExportLogUpdate (Annex ())
+
+instance Semigroup PostExportLogUpdate where
+	PostExportLogUpdate a <> PostExportLogUpdate b =
+		PostExportLogUpdate (a >> b)
+
+noPostExportLogUpdate :: PostExportLogUpdate
+noPostExportLogUpdate = PostExportLogUpdate (return ())
+
+postExportLogUpdate :: PostExportLogUpdate -> Annex ()
+postExportLogUpdate (PostExportLogUpdate a) = a
 
 data Diffed t
 	= DiffChanged t
@@ -538,7 +567,10 @@ importChanges remote importtreeconfig importcontent thirdpartypopulated importab
 					Nothing -> fullimport currcidtree
 					Just lastimportedtree -> diffimport cidtreemap prevcidtree currcidtree lastimportedtree
   where
-	remember = recordContentIdentifierTree (Remote.uuid remote)
+  	-- Record the content identifier tree after the export log is
+	-- updated for the import.
+	remember = PostExportLogUpdate .
+		recordContentIdentifierTree (Remote.uuid remote)
 
 	-- In order to use a diff, the previous ContentIdentifier tree must
 	-- not have been garbage collected. Which can happen since there
@@ -559,11 +591,11 @@ importChanges remote importtreeconfig importcontent thirdpartypopulated importab
 						)
 
 	fullimport currcidtree = 
-		importKeys remote importtreeconfig importcontent thirdpartypopulated importablecontents >>= \case
-			ImportUnfinished -> return ImportUnfinished
-			ImportFinished r -> do
-				remember currcidtree
-	 			return $ ImportFinished $ ImportedFull r
+		importKeys remote importtreeconfig importcontent thirdpartypopulated importablecontents >>= return . \case
+			ImportUnfinished -> ImportUnfinished
+			ImportFinished a r -> 
+				ImportFinished (a <> remember currcidtree) $
+					ImportedFull r
 		
 	diffimport cidtreemap prevcidtree currcidtree lastimportedtree = do
 		(diff, cleanup) <- inRepo $ Git.DiffTree.diffTreeRecursive
@@ -581,17 +613,15 @@ importChanges remote importtreeconfig importcontent thirdpartypopulated importab
 			ImportUnfinished -> do
 				void $ liftIO cleanup
 				return ImportUnfinished
-			ImportFinished (ImportableContentsComplete ic') -> 
-				liftIO cleanup >>= \case
-					False -> return ImportUnfinished
-					True -> do
-						remember currcidtree
-						return $ ImportFinished $ 
-							ImportedDiff lastimportedtree
-								(mkdiff ic' removed)
+			ImportFinished a (ImportableContentsComplete ic') -> 
+				liftIO cleanup >>= return . \case
+					False -> ImportUnfinished
+					True -> ImportFinished (a <> remember currcidtree) $ 
+						ImportedDiff lastimportedtree
+							(mkdiff ic' removed)
 			-- importKeys is not passed ImportableContentsChunked
 			-- above, so it cannot return it
-			ImportFinished (ImportableContentsChunked {}) -> error "internal"
+			ImportFinished _ (ImportableContentsChunked {}) -> error "internal"
 		
 	isremoval ti = Git.DiffTree.dstsha ti `elem` nullShas
 	
@@ -677,12 +707,12 @@ importKeys remote importtreeconfig importcontent thirdpartypopulated importablec
 			ImportableContentsComplete ic ->
 				go False largematcher cidmap importing db ic >>= return . \case
 					Nothing -> ImportUnfinished
-					Just v -> ImportFinished $ ImportableContentsComplete v
+					Just v -> ImportFinished noPostExportLogUpdate $ ImportableContentsComplete v
 			ImportableContentsChunked {} -> do
 				c <- gochunked db (importableContentsChunk importablecontents)
 				gohistory largematcher cidmap importing db (importableHistoryComplete importablecontents) >>= return . \case
 					Nothing -> ImportUnfinished
-					Just h -> ImportFinished $ ImportableContentsChunked
+					Just h -> ImportFinished noPostExportLogUpdate $ ImportableContentsChunked
 						{ importableContentsChunk = c
 						, importableHistoryComplete = h
 						}
@@ -768,7 +798,7 @@ importKeys remote importtreeconfig importcontent thirdpartypopulated importablec
 					Right r -> next $ do
 						liftIO $ atomically $
 							putTMVar job r
-						return True
+						return (isJust r)
 			commandAction $ bracket_
 				(waitstart importing cid)
 				(signaldone importing cid)
@@ -853,10 +883,10 @@ importKeys remote importtreeconfig importcontent thirdpartypopulated importablec
 			let af = AssociatedFile (Just f)
 			let downloader p' tmpfile = do
 				_ <- Remote.retrieveExportWithContentIdentifier
-					ia loc [cid] (fromRawFilePath tmpfile)
+					ia loc [cid] tmpfile
 					(Left k)
 					(combineMeterUpdate p' p)
-				ok <- moveAnnex k af tmpfile
+				ok <- moveAnnex k tmpfile
 				when ok $
 					logStatus NoLiveUpdate k InfoPresent
 				return (Just (k, ok))
@@ -871,7 +901,7 @@ importKeys remote importtreeconfig importcontent thirdpartypopulated importablec
 	doimportsmall cidmap loc cid sz p = do
 		let downloader tmpfile = do
 			(k, _) <- Remote.retrieveExportWithContentIdentifier
-				ia loc [cid] (fromRawFilePath tmpfile)
+				ia loc [cid] tmpfile
 				(Right (mkkey tmpfile))
 				p
 			case keyGitSha k of
@@ -894,12 +924,12 @@ importKeys remote importtreeconfig importcontent thirdpartypopulated importablec
 		let af = AssociatedFile (Just f)
 		let downloader tmpfile p = do
 			(k, _) <- Remote.retrieveExportWithContentIdentifier
-				ia loc [cid] (fromRawFilePath tmpfile)
+				ia loc [cid] tmpfile
 				(Right (mkkey tmpfile))
 				p
 			case keyGitSha k of
 				Nothing -> do
-					ok <- moveAnnex k af tmpfile
+					ok <- moveAnnex k tmpfile
 					when ok $ do
 						recordcidkey cidmap cid k
 						logStatus NoLiveUpdate k InfoPresent
@@ -950,16 +980,25 @@ importKeys remote importtreeconfig importcontent thirdpartypopulated importablec
 		case importtreeconfig of
 			ImportTree -> fromImportLocation loc
 			ImportSubTree subdir _ ->
-				getTopFilePath subdir P.</> fromImportLocation loc
+				getTopFilePath subdir </> fromImportLocation loc
 
-	getcidkey cidmap db cid = liftIO $
+	getcidkey cidmap db cid = do
 		-- Avoiding querying the database when it's empty speeds up
 		-- the initial import.
-		if CIDDb.databaseIsEmpty db
+		l <- liftIO $ if CIDDb.databaseIsEmpty db
 			then getcidkeymap cidmap cid
 			else CIDDb.getContentIdentifierKeys db rs cid >>= \case
 				[] -> getcidkeymap cidmap cid
 				l -> return l
+		filterM validcidkey l
+	
+	-- Guard against a content identifier containing a git sha that is
+	-- not present in the repository. This can happen when a previous,
+	-- import failed and the tree was not recorded, and this import is
+	-- being run in another clone of the repository.
+	validcidkey k = case keyGitSha k of
+		Just sha -> isJust <$> catObjectMetaData sha
+		Nothing -> return True
 
 	getcidkeymap cidmap cid =
 		atomically $ maybeToList . M.lookup cid <$> readTVar cidmap
@@ -1041,6 +1080,10 @@ pruneImportMatcher = Utility.Matcher.pruneMatcher matchNeedsKey
  - write a git tree that contains that, git will complain and refuse to
  - check it out.
  -
+ - Filters out any paths that contain an empty filename, because git cannot
+ - represent an empty filename in a tree, but some special remotes do
+ - support empty filenames.
+ -
  - Filters out new things not matching the FileMatcher or that are
  - gitignored. However, files that are already in git get imported
  - regardless. (Similar to how git add behaves on gitignored files.)
@@ -1087,15 +1130,35 @@ getImportableContents r importtreeconfig ci matcher = do
 
 	wanted dbhandle (loc, (_cid, sz))
 		| ingitdir = pure False
+		| OS.null (fromImportLocation loc) = do
+			warning $ UnquotedString "Cannot import a file with an empty filename"
+			return False
+		| isdirectory = do
+			warning $ UnquotedString "Cannot import a file with a name that appears to be a directory: "
+				<> QuotedPath (fromImportLocation loc)
+			return False
 		| otherwise =
 			isknown <||> (matches <&&> notignored)
 	  where
 		-- Checks, from least to most expensive.
-		ingitdir = ".git" `elem` Posix.splitDirectories (fromImportLocation loc)
+#ifdef mingw32_HOST_OS
+		ingitdir = ".git" `elem` Posix.splitDirectories loc'
+#else
+		ingitdir = literalOsPath ".git" `elem` splitDirectories (fromImportLocation loc)
+#endif
+#ifdef mingw32_HOST_OS
+		isdirectory = Posix.dropFileName loc' == loc'
+#else
+		isdirectory = dropFileName (fromImportLocation loc) == fromImportLocation loc
+#endif
 		matches = matchesImportLocation matcher loc sz
 		isknown = isKnownImportLocation dbhandle loc
 		notignored = notIgnoredImportLocation importtreeconfig ci loc
-	
+
+#ifdef mingw32_HOST_OS
+		loc' = fromOsPath (fromImportLocation loc)
+#endif
+
 	wantedunder dbhandle root (loc, v) = 
 		wanted dbhandle (importableContentsChunkFullLocation root loc, v)
 
@@ -1120,6 +1183,6 @@ notIgnoredImportLocation importtreeconfig ci loc = not <$> checkIgnored ci f
   where
 	f = case importtreeconfig of
 		ImportSubTree dir _ ->
-			getTopFilePath dir P.</> fromImportLocation loc
+			getTopFilePath dir </> fromImportLocation loc
 		ImportTree ->
 			fromImportLocation loc
