@@ -47,7 +47,6 @@ module Annex.Locations (
 	gitAnnexUnusedLog,
 	gitAnnexKeysDbDir,
 	gitAnnexKeysDbLock,
-	gitAnnexKeysDbIndexCache,
 	gitAnnexFsckState,
 	gitAnnexFsckDbDir,
 	gitAnnexFsckDbDirOld,
@@ -411,11 +410,6 @@ gitAnnexKeysDbDir r c =
 gitAnnexKeysDbLock :: Git.Repo -> GitConfig -> OsPath
 gitAnnexKeysDbLock  r c = gitAnnexKeysDbDir r c <> literalOsPath ".lck"
 
-{- Contains the stat of the last index file that was
- - reconciled with the keys database. -}
-gitAnnexKeysDbIndexCache :: Git.Repo -> GitConfig -> OsPath
-gitAnnexKeysDbIndexCache r c = gitAnnexKeysDbDir r c <> literalOsPath ".cache"
-
 {- .git/annex/fsck/uuid/ is used to store information about incremental
  - fscks. -}
 gitAnnexFsckDir :: UUID -> Git.Repo -> Maybe GitConfig -> OsPath
@@ -802,11 +796,7 @@ keyFile k =
 
 	needesc = map (fromIntegral . ord) ['&', '%', ':', '/']
 
-#if MIN_VERSION_bytestring(0,11,3)
 	anyneedesc = SB.any (`elem` needesc)
-#else
-	anyneedesc = any (`elem` needesc) . SB.unpack
-#endif
 
 {- Reverses keyFile, converting a filename fragment (ie, the basename of
  - the symlink target) into a key. -}
