@@ -1,6 +1,6 @@
 {- git-annex command
  -
- - Copyright 2010-2021 Joey Hess <id@joeyh.name>
+ - Copyright 2010-2025 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
@@ -20,6 +20,7 @@ import Annex.NumCopies
 import Annex.Content
 import Annex.Wanted
 import Annex.Notification
+import Remote.List
 
 cmd :: Command
 cmd = withAnnexOptions [jobsOption, jsonOptions, annexedMatchingOptions] $
@@ -60,7 +61,7 @@ seek o = startConcurrency commandStages $ do
 				then pure Nothing
 				else pure (Just remote)				
 	let seeker = AnnexedFileSeeker
-		{ startAction = start o from
+		{ startAction = const $ start o from
 		, checkContentPresent = case from of
 			Nothing -> Just True
 			Just _ -> Nothing
@@ -76,22 +77,24 @@ seek o = startConcurrency commandStages $ do
   where
 	ww = WarnUnmatchLsFiles "drop"
 
-start :: DropOptions -> Maybe Remote -> SeekInput -> RawFilePath -> Key -> CommandStart
+start :: DropOptions -> Maybe Remote -> SeekInput -> OsPath -> Key -> CommandStart
 start o from si file key = start' o from key afile ai si
   where
 	afile = AssociatedFile (Just file)
 	ai = mkActionItem (key, afile)
 
 start' :: DropOptions -> Maybe Remote -> Key -> AssociatedFile -> ActionItem -> SeekInput -> CommandStart
-start' o from key afile ai si = 
-	checkDropAuto (autoMode o) from afile key $ \numcopies mincopies ->
-		stopUnless wantdrop $
+start' o from key afile ai si = do
+	checkDropAuto (autoMode o) from afile key $ \numcopies mincopies -> do
+		lu <- prepareLiveUpdate remoteuuid key RemovingKey
+		stopUnless (wantdrop lu) $
 			case from of
-				Nothing -> startLocal pcc afile ai si numcopies mincopies key [] ud
-				Just remote -> startRemote pcc afile ai si numcopies mincopies key ud remote
+				Nothing -> startLocal lu pcc afile ai si numcopies mincopies key [] ud
+				Just remote -> startRemote lu pcc afile ai si numcopies mincopies key ud remote
   where
-	wantdrop
-		| autoMode o = wantDrop False (Remote.uuid <$> from) (Just key) afile Nothing
+	remoteuuid = Remote.uuid <$> from
+	wantdrop lu
+		| autoMode o = wantDrop lu False remoteuuid (Just key) afile Nothing
 		| otherwise = return True
 	pcc = PreferredContentChecked (autoMode o)
 	ud = case (batchOption o, keyOptions o) of
@@ -101,31 +104,40 @@ start' o from key afile ai si =
 startKeys :: DropOptions -> Maybe Remote -> (SeekInput, Key, ActionItem) -> CommandStart
 startKeys o from (si, key, ai) = start' o from key (AssociatedFile Nothing) ai si
 
-startLocal :: PreferredContentChecked -> AssociatedFile -> ActionItem -> SeekInput -> NumCopies -> MinCopies -> Key -> [VerifiedCopy] -> DroppingUnused -> CommandStart
-startLocal pcc afile ai si numcopies mincopies key preverified ud =
+startLocal :: LiveUpdate -> PreferredContentChecked -> AssociatedFile -> ActionItem -> SeekInput -> NumCopies -> MinCopies -> Key -> [VerifiedCopy] -> DroppingUnused -> CommandStart
+startLocal lu pcc afile ai si numcopies mincopies key preverified ud =
 	starting "drop" (OnlyActionOn key ai) si $
-		performLocal pcc key afile numcopies mincopies preverified ud
+		performLocal lu pcc key afile numcopies mincopies preverified ud
 
-startRemote :: PreferredContentChecked -> AssociatedFile -> ActionItem -> SeekInput -> NumCopies -> MinCopies -> Key -> DroppingUnused -> Remote -> CommandStart
-startRemote pcc afile ai si numcopies mincopies key ud remote = 
-	starting "drop" (OnlyActionOn key ai) si $ do
+startRemote :: LiveUpdate -> PreferredContentChecked -> AssociatedFile -> ActionItem -> SeekInput -> NumCopies -> MinCopies -> Key -> DroppingUnused -> Remote -> CommandStart
+startRemote lu pcc afile ai si numcopies mincopies key ud remote = do
+	fast <- Annex.getRead Annex.fast
+	if fast
+		then do
+			remotes <- Remote.keyPossibilities (Remote.IncludeIgnored True) key
+			if remote `elem` remotes
+				then go
+				else stop
+		else go
+  where
+	go = starting "drop" (OnlyActionOn key ai) si $ do
 		showAction $ UnquotedString $ "from " ++ Remote.name remote
-		performRemote pcc key afile numcopies mincopies remote ud
+		performRemote lu pcc key afile numcopies mincopies remote ud
 
-performLocal :: PreferredContentChecked -> Key -> AssociatedFile -> NumCopies -> MinCopies -> [VerifiedCopy] -> DroppingUnused -> CommandPerform
-performLocal pcc key afile numcopies mincopies preverified ud = lockContentForRemoval key fallback $ \contentlock -> do
+performLocal :: LiveUpdate -> PreferredContentChecked -> Key -> AssociatedFile -> NumCopies -> MinCopies -> [VerifiedCopy] -> DroppingUnused -> CommandPerform
+performLocal lu pcc key afile numcopies mincopies preverified ud = lockContentForRemoval key fallback $ \contentlock -> do
 	u <- getUUID
 	(tocheck, verified) <- verifiableCopies key [u]
-	doDrop pcc u (Just contentlock) key afile numcopies mincopies [] (preverified ++ verified) tocheck
+	doDrop lu pcc u (Just contentlock) key afile numcopies mincopies [] (preverified ++ verified) tocheck
 		( \proof -> do
 			fastDebug "Command.Drop" $ unwords
 				[ "Dropping from here"
 				, "proof:"
 				, show proof
 				]
-			removeAnnex contentlock
+			removeAnnex remoteList contentlock
 			notifyDrop afile True
-			next $ cleanupLocal key ud
+			next $ cleanupLocal lu key ud
 		, do 
 			notifyDrop afile False
 			stop
@@ -136,14 +148,14 @@ performLocal pcc key afile numcopies mincopies preverified ud = lockContentForRe
 	-- is present, but due to buffering, may find it present for the
 	-- second file before the first is dropped. If so, nothing remains
 	-- to be done except for cleaning up.
-	fallback = next $ cleanupLocal key ud
+	fallback = next $ cleanupLocal lu key ud
 
-performRemote :: PreferredContentChecked -> Key -> AssociatedFile -> NumCopies -> MinCopies -> Remote -> DroppingUnused -> CommandPerform
-performRemote pcc key afile numcopies mincopies remote ud = do
+performRemote :: LiveUpdate -> PreferredContentChecked -> Key -> AssociatedFile -> NumCopies -> MinCopies -> Remote -> DroppingUnused -> CommandPerform
+performRemote lu pcc key afile numcopies mincopies remote ud = do
 	-- Filter the uuid it's being dropped from out of the lists of
 	-- places assumed to have the key, and places to check.
 	(tocheck, verified) <- verifiableCopies key [uuid]
-	doDrop pcc uuid Nothing key afile numcopies mincopies [uuid] verified tocheck
+	doDrop lu pcc uuid Nothing key afile numcopies mincopies [uuid] verified tocheck
 		( \proof -> do 
 			fastDebug "Command.Drop" $ unwords
 				[ "Dropping from remote"
@@ -151,22 +163,22 @@ performRemote pcc key afile numcopies mincopies remote ud = do
 				, "proof:"
 				, show proof
 				]
-			ok <- Remote.action (Remote.removeKey remote key)
-			next $ cleanupRemote key remote ud ok
+			ok <- Remote.action (Remote.removeKey remote proof key)
+			next $ cleanupRemote lu key remote ud ok
 		, stop
 		)
   where
 	uuid = Remote.uuid remote
 
-cleanupLocal :: Key -> DroppingUnused -> CommandCleanup
-cleanupLocal key ud = do
-	logStatus key (dropStatus ud)
+cleanupLocal :: LiveUpdate -> Key -> DroppingUnused -> CommandCleanup
+cleanupLocal lu key ud = do
+	logStatus lu key (dropStatus ud)
 	return True
 
-cleanupRemote :: Key -> Remote -> DroppingUnused -> Bool -> CommandCleanup
-cleanupRemote key remote ud ok = do
+cleanupRemote :: LiveUpdate -> Key -> Remote -> DroppingUnused -> Bool -> CommandCleanup
+cleanupRemote lu key remote ud ok = do
 	when ok $
-		Remote.logStatus remote key (dropStatus ud)
+		Remote.logStatus lu remote key (dropStatus ud)
 	return ok
 
 {- Set when the user explicitly chose to operate on unused content.
@@ -189,7 +201,8 @@ dropStatus (DroppingUnused True) = InfoDead
  - --force overrides and always allows dropping.
  -}
 doDrop
-	:: PreferredContentChecked
+	:: LiveUpdate
+	-> PreferredContentChecked
 	-> UUID
 	-> Maybe ContentRemovalLock
 	-> Key
@@ -201,11 +214,11 @@ doDrop
 	-> [UnVerifiedCopy]
 	-> (Maybe SafeDropProof -> CommandPerform, CommandPerform)
 	-> CommandPerform
-doDrop pcc dropfrom contentlock key afile numcopies mincopies skip preverified check (dropaction, nodropaction) = 
+doDrop lu pcc dropfrom contentlock key afile numcopies mincopies skip preverified check (dropaction, nodropaction) = 
 	ifM (Annex.getRead Annex.force)
 		( dropaction Nothing
-		, ifM (checkRequiredContent pcc dropfrom key afile)
-			( verifyEnoughCopiesToDrop nolocmsg key 
+		, ifM (checkRequiredContent lu pcc dropfrom key afile)
+			( verifyEnoughCopiesToDrop nolocmsg key (Just dropfrom)
 				contentlock numcopies mincopies
 				skip preverified check
 					(dropaction . Just)
@@ -225,10 +238,10 @@ doDrop pcc dropfrom contentlock key afile numcopies mincopies skip preverified c
  - providing this avoids that extra work. -}
 newtype PreferredContentChecked = PreferredContentChecked Bool
 
-checkRequiredContent :: PreferredContentChecked -> UUID -> Key -> AssociatedFile -> Annex Bool
-checkRequiredContent (PreferredContentChecked True) _ _ _ = return True
-checkRequiredContent (PreferredContentChecked False) u k afile =
-	checkDrop isRequiredContent False (Just u) (Just k) afile Nothing >>= \case
+checkRequiredContent :: LiveUpdate -> PreferredContentChecked -> UUID -> Key -> AssociatedFile -> Annex Bool
+checkRequiredContent _ (PreferredContentChecked True) _ _ _ = return True
+checkRequiredContent lu (PreferredContentChecked False) u k afile =
+	checkDrop isRequiredContent lu False (Just u) (Just k) afile Nothing >>= \case
 		Nothing -> return True
 		Just afile' -> do
 			if afile == afile'
@@ -253,7 +266,7 @@ checkDropAuto automode mremote afile key a =
 			uuid <- getUUID
 			let remoteuuid = fromMaybe uuid $ Remote.uuid <$> mremote
 			locs' <- trustExclude UnTrusted $ filter (/= remoteuuid) locs
-			if length locs' >= fromNumCopies numcopies
+			if numCopiesCheck'' locs' (>=) numcopies
 				then a numcopies mincopies
 				else stop
 		| otherwise = a numcopies mincopies
