@@ -34,7 +34,6 @@ import Types.WorkerPool
 import Remote.List
 import Types.Direction
 import Utility.Metered
-import Utility.STM
 
 import Servant
 import qualified Servant.Types.SourceT as S
@@ -42,6 +41,7 @@ import qualified Data.ByteString as B
 import qualified Data.ByteString.Lazy as L
 import qualified Data.ByteString.Lazy.Internal as LI
 import Control.Concurrent.Async
+import Control.Concurrent.STM
 import Control.Concurrent
 import System.IO.Unsafe
 import Data.Either
