@@ -1,6 +1,6 @@
 {- git-annex remotes
  -
- - Copyright 2011-2020 Joey Hess <id@joeyh.name>
+ - Copyright 2011-2024 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
@@ -26,7 +26,6 @@ module Remote (
 	remoteTypes,
 	remoteList,
 	remoteList',
-	gitSyncableRemoteType,
 	remoteMap,
 	remoteMap',
 	uuidDescriptions,
@@ -41,16 +40,19 @@ module Remote (
 	prettyPrintUUIDs,
 	prettyPrintUUIDsDescs,
 	prettyPrintUUIDsWith,
+	prettyPrintUUIDsWith',
 	prettyListUUIDs,
 	prettyUUID,
 	remoteFromUUID,
 	remotesWithUUID,
 	remotesWithoutUUID,
 	keyLocations,
+	IncludeIgnored(..),
 	keyPossibilities,
 	remoteLocations,
 	nameToUUID,
 	nameToUUID',
+	nameToUUID'',
 	showTriedRemotes,
 	listRemoteNames,
 	showLocations,
@@ -58,7 +60,10 @@ module Remote (
 	logStatus,
 	checkAvailable,
 	claimingUrl,
+	claimingUrl',
 	isExportSupported,
+	gitSyncableRemote,
+	gitSyncableRemoteType,
 ) where
 
 import Data.Ord
@@ -80,7 +85,7 @@ import Remote.List
 import Remote.List.Util
 import Config
 import Config.DynamicConfig
-import Git.Types (RemoteName, ConfigKey(..), fromConfigValue)
+import Git.Types (ConfigKey(..), fromConfigValue)
 import Utility.Aeson
 
 {- Map from UUIDs of Remotes to a calculated value. -}
@@ -144,8 +149,11 @@ byNameWithUUID = checkuuid <=< byName
 		| otherwise = return $ Just r
 
 byName' :: RemoteName -> Annex (Either String Remote)
-byName' "" = return $ Left "no repository name specified"
-byName' n = go . filter matching <$> remoteList
+byName' n = byName'' n <$> remoteList
+
+byName'' :: RemoteName -> [Remote] -> Either String Remote
+byName'' "" _ = Left "no repository name specified"
+byName'' n remotelist = go $ filter matching remotelist
   where
 	go [] = Left $ "there is no available git remote named \"" ++ n ++ "\""
 	go (match:_) = Right match
@@ -178,20 +186,31 @@ nameToUUID n = nameToUUID' n >>= \case
 	(_, msg) -> giveup msg
 
 nameToUUID' :: RemoteName -> Annex ([UUID], String)
-nameToUUID' n
+nameToUUID' n = do
+	f <- nameToUUID''
+	return (f n)
+
+nameToUUID'' :: Annex (RemoteName -> ([UUID], String))
+nameToUUID'' = do
+	l <- remoteList
+	u <- getUUID
+	m <- uuidDescMap
+	return $ \n -> nameToUUID''' n l u m
+
+nameToUUID''' :: RemoteName -> [Remote] -> UUID -> UUIDDescMap -> ([UUID], String)
+nameToUUID''' n remotelist hereu m
 	| n == "." = currentrepo
 	| n == "here" = currentrepo
-	| otherwise = byName' n >>= go
+	| otherwise = go (byName'' n remotelist)
   where
-	currentrepo = mkone <$> getUUID
+	currentrepo = mkone hereu
 
-	go (Right r) = return $ case uuid r of
+	go (Right r) = case uuid r of
 		NoUUID -> ([], noRemoteUUIDMsg r)
 		u -> mkone u
-	go (Left e) = do
-		m <- uuidDescMap
+	go (Left e) =
 		let descn = UUIDDesc (encodeBS n)
-		return $ case M.keys (M.filter (== descn) m) of
+		in case M.keys (M.filter (== descn) m) of
 			[] -> 
 				let u = toUUID n
 				in case M.keys (M.filterWithKey (\k _ -> k == u) m) of
@@ -226,11 +245,25 @@ prettyPrintUUIDsWith
 	-> (v -> Maybe String)
 	-> [(UUID, Maybe v)] 
 	-> Annex String
-prettyPrintUUIDsWith optfield header descm showval uuidvals = do
+prettyPrintUUIDsWith = prettyPrintUUIDsWith' True
+
+prettyPrintUUIDsWith'
+	:: ToJSON' v
+	=> Bool
+	-> Maybe String 
+	-> String 
+	-> UUIDDescMap
+	-> (v -> Maybe String)
+	-> [(UUID, Maybe v)] 
+	-> Annex String
+prettyPrintUUIDsWith' indented optfield header descm showval uuidvals = do
 	hereu <- getUUID
 	maybeShowJSON $ JSONChunk [(header, V.fromList $ map (jsonify hereu) uuidvals)]
-	return $ unwords $ map (\u -> "\t" ++ prettify hereu u ++ "\n") uuidvals
+	return $ concatMap 
+		(\u -> tabindent ++ prettify hereu u ++ "\n")
+		uuidvals
   where
+	tabindent = if indented then "\t" else ""
 	finddescription u = fromUUIDDesc $ M.findWithDefault mempty u descm
 	prettify hereu (u, optval)
 		| not (null d) = addoptval $ fromUUID u ++ " -- " ++ d
@@ -244,7 +277,7 @@ prettyPrintUUIDsWith optfield header descm showval uuidvals = do
 			| otherwise = n
 		addoptval s = case showval =<< optval of
 			Nothing -> s
-			Just val -> val ++ ": " ++ s
+			Just val -> val ++ s
 	jsonify hereu (u, optval) = object $ catMaybes
 		[ Just ("uuid", toJSON' (fromUUID u :: String))
 		, Just ("description", toJSON' $ finddescription u)
@@ -286,14 +319,6 @@ remoteFromUUID u = ifM ((==) u <$> getUUID)
 		remotesChanged
 		findinmap
 
-{- Filters a list of remotes to ones that have the listed uuids. -}
-remotesWithUUID :: [Remote] -> [UUID] -> [Remote]
-remotesWithUUID rs us = filter (\r -> uuid r `elem` us) rs
-
-{- Filters a list of remotes to ones that do not have the listed uuids. -}
-remotesWithoutUUID :: [Remote] -> [UUID] -> [Remote]
-remotesWithoutUUID rs us = filter (\r -> uuid r `notElem` us) rs
-
 {- List of repository UUIDs that the location log indicates may have a key.
  - Dead repositories are excluded. -}
 keyLocations :: Key -> Annex [UUID]
@@ -304,41 +329,27 @@ keyLocations key = trustExclude DeadTrusted =<< loggedLocations key
  -
  - Also includes remotes with remoteAnnexSpeculatePresent set.
  -}
-keyPossibilities :: Key -> Annex [Remote]
-keyPossibilities key = do
-	u <- getUUID
-	-- uuids of all remotes that are recorded to have the key
-	locations <- filter (/= u) <$> keyLocations key
-	speclocations <- map uuid
-		. filter (remoteAnnexSpeculatePresent . gitconfig)
-		<$> remoteList
-	-- there are unlikely to be many speclocations, so building a Set
-	-- is not worth the expense
-	let locations' = speclocations ++ filter (`notElem` speclocations) locations
-	fst <$> remoteLocations locations' []
+keyPossibilities :: IncludeIgnored -> Key -> Annex [Remote]
+keyPossibilities ii key = do
+	locations <- keyLocations key
+	keyPossibilities' ii key locations =<< remoteList
 
 {- Given a list of locations of a key, and a list of all
  - trusted repositories, generates a cost-ordered list of
  - remotes that contain the key, and a list of trusted locations of the key.
  -}
-remoteLocations :: [UUID] -> [UUID] -> Annex ([Remote], [UUID])
-remoteLocations locations trusted = do
-	let validtrustedlocations = nub locations `intersect` trusted
-
-	-- remotes that match uuids that have the key
-	allremotes <- remoteList 
-		>>= filterM (not <$$> liftIO . getDynamicConfig . remoteAnnexIgnore . gitconfig)
-	let validremotes = remotesWithUUID allremotes locations
-
-	return (sortBy (comparing cost) validremotes, validtrustedlocations)
+remoteLocations :: IncludeIgnored -> [UUID] -> [UUID] -> Annex ([Remote], [UUID])
+remoteLocations ii locations trusted = 
+	remoteLocations' ii locations trusted =<< remoteList
 
 {- Displays known locations of a key and helps the user take action
  - to make them accessible. -}
-showLocations :: Bool -> Key -> [UUID] -> String -> Annex ()
-showLocations separateuntrusted key exclude nolocmsg = do
+showLocations :: Bool -> Key -> (UUID -> Annex Bool) -> String -> Annex ()
+showLocations separateuntrusted key checkexclude nolocmsg = do
 	u <- getUUID
 	remotes <- remoteList
 	uuids <- keyLocations key
+	exclude <- filterM checkexclude uuids
 	untrusteduuids <- if separateuntrusted
 		then trustGet UnTrusted
 		else pure []
@@ -401,8 +412,8 @@ forceTrust level remotename = do
  - in the local repo, not on the remote. The process of transferring the
  - key to the remote, or removing the key from it *may* log the change
  - on the remote, but this cannot always be relied on. -}
-logStatus :: Remote -> Key -> LogStatus -> Annex ()
-logStatus remote key = logChange key (uuid remote)
+logStatus :: LiveUpdate -> Remote -> Key -> LogStatus -> Annex ()
+logStatus lu remote key = logChange lu key (uuid remote)
 
 {- Orders remotes by cost, with ones with the lowest cost grouped together. -}
 byCost :: [Remote] -> [[Remote]]
@@ -411,9 +422,12 @@ byCost = map snd . sortBy (comparing fst) . M.toList . costmap
 	costmap = M.fromListWith (++) . map costpair
 	costpair r = (cost r, [r])
 
-checkAvailable :: Bool -> Remote -> IO Bool
-checkAvailable assumenetworkavailable = 
-	maybe (return assumenetworkavailable) doesDirectoryExist . localpath
+checkAvailable :: Bool -> Remote -> Annex Bool
+checkAvailable assumenetworkavailable r = tryNonAsync (availability r) >>= \case
+	Left _e -> return assumenetworkavailable
+	Right LocallyAvailable -> return True
+	Right GloballyAvailable -> return assumenetworkavailable
+	Right Unavailable -> return False
 
 hasKey :: Remote -> Key -> Annex (Either String Bool)
 hasKey r k = either (Left  . show) Right <$> tryNonAsync (checkPresent r k)
@@ -423,9 +437,28 @@ hasKeyCheap = checkPresentCheap
 
 {- The web special remote claims urls by default. -}
 claimingUrl :: URLString -> Annex Remote
-claimingUrl url = do
+claimingUrl = claimingUrl' (const True)
+
+{- The web special remote still claims urls if there is no
+ - other remote that does, even when the remotefilter does
+ - not include it. -}
+claimingUrl' :: (Remote -> Bool) -> URLString -> Annex Remote
+claimingUrl' remotefilter url = do
 	rs <- remoteList
-	let web = Prelude.head $ filter (\r -> uuid r == webUUID) rs
-	fromMaybe web <$> firstM checkclaim rs
+	let web = fromMaybe (error "internal") $ headMaybe $
+		filter (\r -> uuid r == webUUID) rs
+	fromMaybe web <$> firstM checkclaim (filter remotefilter rs)
   where
 	checkclaim = maybe (pure False) (`id` url) . claimUrl
+
+{- Is this a remote of a type that git pull and push work with?
+ - That includes special remotes with an annex:: url configured.
+ - It does not include proxied remotes. -}
+gitSyncableRemote :: Remote -> Bool
+gitSyncableRemote r
+	| gitSyncableRemoteType (remotetype r) 
+		&& isJust (remoteUrl (gitconfig r)) =
+			not (isJust (remoteAnnexProxiedBy (gitconfig r)))
+	| otherwise = case remoteUrl (gitconfig r) of
+		Just u | "annex::" `isPrefixOf` u -> True
+		_ -> False
