@@ -1,25 +1,28 @@
 {- Helper to make remotes support export and import (or not).
  -
- - Copyright 2017-2019 Joey Hess <id@joeyh.name>
+ - Copyright 2017-2024 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
 
 {-# LANGUAGE FlexibleInstances, FlexibleContexts #-}
+{-# LANGUAGE OverloadedStrings #-}
 
 module Remote.Helper.ExportImport where
 
 import Annex.Common
+import qualified Annex
 import Types.Remote
 import Types.Key
 import Types.ProposedAccepted
-import Annex.Verify
 import Remote.Helper.Encryptable (encryptionIsEnabled)
 import qualified Database.Export as Export
 import qualified Database.ContentIdentifier as ContentIdentifier
 import Annex.Export
 import Annex.LockFile
 import Annex.SpecialRemote.Config
+import Annex.Verify
+import Annex.Content
 import Git.Types (fromRef)
 import Logs.Export
 import Logs.ContentIdentifier (recordContentIdentifier)
@@ -39,9 +42,8 @@ instance HasExportUnsupported (ExportActions Annex) where
 		, retrieveExport = nope
 		, checkPresentExport = \_ _ -> return False
 		, removeExport = nope
-		, versionedExport = False
 		, removeExportDirectory = nope
-		, renameExport = \_ _ _ -> return Nothing
+		, renameExport = Nothing
 		}
 	 where
 	 	nope = giveup "export not supported"
@@ -77,7 +79,7 @@ importIsSupported = \_ _ -> return True
 adjustExportImportRemoteType :: RemoteType -> RemoteType
 adjustExportImportRemoteType rt = rt { setup = setup' }
   where
-	setup' st mu cp c gc = do
+	setup' st mu remotename cp c gc = do
 		pc <- either giveup return . parseRemoteConfig c
 			=<< configParser rt c
 		let checkconfig supported configured configfield cont =
@@ -95,7 +97,7 @@ adjustExportImportRemoteType rt = rt { setup = setup' }
 				)
 		checkconfig exportSupported exportTree exportTreeField $
 			checkconfig importSupported importTree importTreeField $
-				setup rt st mu cp c gc
+				setup rt st mu remotename cp c gc
 	
 	enable oldc pc configured configfield cont = do
 		oldpc <- parsedRemoteConfig rt oldc
@@ -123,15 +125,17 @@ adjustExportImport r rs = do
 				else importUnsupported
 			}
 		}
+	let annexobjects = isexport && annexObjects (config r)
 	if not isexport && not isimport
 		then return r'
-		else adjustExportImport' isexport isimport r' rs
+		else do
+			gc <- Annex.getGitConfig
+			adjustExportImport' isexport isimport annexobjects r' rs gc
 
-adjustExportImport' :: Bool -> Bool -> Remote -> RemoteStateHandle -> Annex Remote
-adjustExportImport' isexport isimport r rs = do
+adjustExportImport' :: Bool -> Bool -> Bool -> Remote -> RemoteStateHandle -> GitConfig -> Annex Remote
+adjustExportImport' isexport isimport annexobjects r rs gc = do
 	dbv <- prepdbv
 	ciddbv <- prepciddb
-	let versioned = versionedExport (exportActions r)
 	return $ r
 		{ exportActions = if isexport
 			then if isimport
@@ -141,52 +145,55 @@ adjustExportImport' isexport isimport r rs = do
 		, importActions = if isimport
 			then importActions r
 			else importUnsupported
-		, storeKey = \k af p ->
-			-- Storing a key on an export could be implemented,
-			-- but it would perform unncessary work
+		, storeKey = \k af o p ->
+			-- Storing a key to an export location could be
+			-- implemented, but it would perform unnecessary work
 			-- when another repository has already stored the
 			-- key, and the local repository does not know
 			-- about it. To avoid unnecessary costs, don't do it.
 			if thirdpartypopulated
 				then giveup "remote is not populated by git-annex"
 				else if isexport
-					then giveup "remote is configured with exporttree=yes; use `git-annex export` to store content on it"
+					then if annexobjects
+						then storeannexobject k o p
+						else giveup "remote is configured with exporttree=yes; use `git-annex export` to store content on it"
 					else if isimport
 						then giveup "remote is configured with importtree=yes and without exporttree=yes; cannot modify content stored on it"
-						else storeKey r k af p
-		, removeKey = \k -> 
-			-- Removing a key from an export would need to
-			-- change the tree in the export log to not include
+						else storeKey r k af o p
+		, removeKey = \proof k -> 
+			-- Removing a key from an export location would need
+			-- to change the tree in the export log to not include
 			-- the file. Otherwise, conflicts when removing
 			-- files would not be dealt with correctly.
 			-- There does not seem to be a good use case for
-			-- removing a key from an export in any case.
+			-- removing a key from an exported tree in any case.
 			if thirdpartypopulated
 				then giveup "dropping content from this remote is not supported"
 				else if isexport
-					then giveup "dropping content from an export is not supported; use `git annex export` to export a tree that lacks the files you want to remove"
+					then if annexobjects
+						then removeannexobject dbv k
+						else giveup "dropping content from an export is not supported; use `git annex export` to export a tree that lacks the files you want to remove"
 					else if isimport
 						then giveup "dropping content from this remote is not supported because it is configured with importtree=yes"
-						else removeKey r k
+						else removeKey r proof k
 		, lockContent = if versioned
 			then lockContent r
 			else Nothing
 		, retrieveKeyFile = \k af dest p vc ->
-			if isimport
+			if isimport || isexport
 				then supportversionedretrieve k af dest p vc $
-					retrieveKeyFileFromImport dbv ciddbv k af dest p
-				else if isexport
-					then supportversionedretrieve k af dest p vc $
-						retrieveKeyFileFromExport dbv k af dest p
-					else retrieveKeyFile r k af dest p vc
+					supportretrieveannexobject dbv k af dest p $
+						retrieveFromImportOrExport (tryexportlocs dbv k) ciddbv k af dest p
+				else retrieveKeyFile r k af dest p vc
 		, retrieveKeyFileCheap = if versioned
 			then retrieveKeyFileCheap r
 			else Nothing
 		, checkPresent = \k -> if versioned
 			then checkPresent r k
 			else if isimport
-				then anyM (checkPresentImport ciddbv k)
-					=<< getanyexportlocs dbv k
+				then checkpresentwith k $
+					anyM (checkPresentImport ciddbv k)
+						=<< getanyexportlocs dbv k
 				else if isexport
 					-- Check if any of the files a key
 					-- was exported to are present. This 
@@ -197,8 +204,9 @@ adjustExportImport' isexport isimport r rs = do
 					-- to it. Remotes that have such 
 					-- problems are made untrusted,
 					-- so it's not worried about here.
-					then anyM (checkPresentExport (exportActions r) k)
-						=<< getanyexportlocs dbv k
+					then checkpresentwith k $
+						anyM (checkPresentExport (exportActions r) k)
+							=<< getanyexportlocs dbv k
 					else checkPresent r k
 		-- checkPresent from an export is more expensive
 		-- than otherwise, so not cheap. Also, this
@@ -212,7 +220,7 @@ adjustExportImport' isexport isimport r rs = do
 		-- ones though, since they still allow accessing by Key.
 		-- And for thirdPartyPopulated, it depends on how the
 		-- content gets actually stored in the remote, so
-		-- is not overriddden here.
+		-- is not overridden here.
 		, untrustworthy =
 			if versioned || thirdPartyPopulated (remotetype r)
 				then untrustworthy r
@@ -226,16 +234,24 @@ adjustExportImport' isexport isimport r rs = do
 				then do
 					ts <- map fromRef . exportedTreeishes
 						<$> getExport (uuid r)
-					return (is++[("exporttree", "yes"), ("exportedtree", unwords ts)])
+					return $ is ++ catMaybes
+						[ Just ("exporttree", "yes")
+						, Just ("exportedtree", unwords ts)
+						, if annexobjects
+							then Just ("annexobjects", "yes")
+							else Nothing
+						]
 				else return is
 			return $ if isimport && not thirdpartypopulated
 				then (is'++[("importtree", "yes")])
 				else is'
 		}
   where
+	versioned = isVersioning (config r)
+	
 	thirdpartypopulated = thirdPartyPopulated (remotetype r)
 
-	-- exportActions adjusted to use the equivilant import actions,
+	-- exportActions adjusted to use the equivalent import actions,
 	-- which take ContentIdentifiers into account.
 	exportActionsForImport dbv ciddbv ea = ea
   		{ storeExport = \f k loc p -> do
@@ -245,7 +261,8 @@ adjustExportImport' isexport isimport r rs = do
 			oldcids <- liftIO $ concat
 				<$> mapM (ContentIdentifier.getContentIdentifiers db rs) oldks
 			newcid <- storeExportWithContentIdentifier (importActions r) f k loc oldcids p
-			withExclusiveLock gitAnnexContentIdentifierLock $ do
+			cidlck <- calcRepo' gitAnnexContentIdentifierLock
+			withExclusiveLock cidlck $ do
 				liftIO $ ContentIdentifier.recordContentIdentifier db rs newcid k
 				liftIO $ ContentIdentifier.flushDbQueue db
 			recordContentIdentifier rs newcid k
@@ -256,7 +273,7 @@ adjustExportImport' isexport isimport r rs = do
 		-- renameExport is optional, and the remote's
 		-- implementation may lose modifications to the file
 		-- (by eg copying and then deleting) so don't use it
-		, renameExport = \_ _ _ -> return Nothing
+		, renameExport = Nothing
 		, checkPresentExport = checkPresentImport ciddbv
 		}
 	
@@ -279,12 +296,14 @@ adjustExportImport' isexport isimport r rs = do
 			Nothing -> ifM (liftIO $ atomically $ tryPutTMVar lcklckv ())
 				( do
 					db <- ContentIdentifier.openDb
-					ContentIdentifier.needsUpdateFromLog db >>= \case
-						Just v -> withExclusiveLock gitAnnexContentIdentifierLock $
-							ContentIdentifier.updateFromLog db v
-						Nothing -> noop
-					liftIO $ atomically $ putTMVar dbtv db
-					return db
+					db' <- ContentIdentifier.needsUpdateFromLog db >>= \case
+						Just v -> do
+							cidlck <- calcRepo' gitAnnexContentIdentifierLock 
+							withExclusiveLock cidlck $
+								ContentIdentifier.updateFromLog db v
+						Nothing -> pure db
+					liftIO $ atomically $ putTMVar dbtv db'
+					return db'
 				-- loser waits for winner to open the db and
 				-- can then also use its handle
 				, liftIO $ atomically (readTMVar dbtv)
@@ -310,7 +329,7 @@ adjustExportImport' isexport isimport r rs = do
 				, liftIO $ atomically (readTMVar dbv)
 				)
 
-	getexportinconflict (_, _, v) = v
+	isexportinconflict (_, _, v) = liftIO $ atomically $ readTVar v
 
 	updateexportdb db exportinconflict =
 		Export.updateExportTreeFromLog db >>= \case
@@ -324,60 +343,126 @@ adjustExportImport' isexport isimport r rs = do
 		db <- getexportdb dbv
 		liftIO $ Export.getExportTree db k
 	
-	getfirstexportloc dbv k = do
-		getexportlocs dbv k >>= \case
-			[] -> giveup "unknown export location"
-			(l:_) -> return l
-	
 	getexportlocs dbv k = do
 		db <- getexportdb dbv
-		liftIO $ Export.getExportTree db k >>= \case
-			[] -> ifM (atomically $ readTVar $ getexportinconflict dbv)
+		liftIO (Export.getExportTree db k) >>= \case
+			[] -> ifM (isexportinconflict dbv)
 				( giveup "unknown export location, likely due to the export conflict"
 				, return []
 				)
 			ls -> return ls
+	
+	tryexportlocs dbv k a = 
+		go Nothing =<< getexportlocs dbv k
+	  where
+		go Nothing [] = giveup "unknown export location"
+		go (Just ex) [] = throwM ex
+		go _ (l:ls) = tryNonAsync (a l) >>= \case
+			Right v -> return v
+			Left e -> go (Just e) ls
 		
 	getkeycids ciddbv k = do
 		db <- getciddb ciddbv
 		liftIO $ ContentIdentifier.getContentIdentifiers db rs k
 
+	retrieveFromImportOrExport getlocs ciddbv k af dest p
+		| isimport = retrieveFromImport getlocs ciddbv k af dest p
+		| otherwise = retrieveFromExport getlocs k af dest p
+
 	-- Keys can be retrieved using retrieveExport, but since that
 	-- retrieves from a path in the remote that another writer could
 	-- have replaced with content not of the requested key, the content
 	-- has to be strongly verified.
-	retrieveKeyFileFromExport dbv k _af dest p = ifM (isVerifiable k)
-		( do
-			l <- getfirstexportloc dbv k
-			retrieveExport (exportActions r) k l dest p >>= return . \case
+	retrieveFromExport getlocs k _af dest p = ifM (isVerifiable k)
+		( getlocs $ \loc -> 
+			retrieveExport (exportActions r) k loc dest p >>= return . \case
 				UnVerified -> MustVerify
 				IncompleteVerify iv -> MustFinishIncompleteVerify iv
 				v -> v
 		, giveup $ "exported content cannot be verified due to using the " ++ decodeBS (formatKeyVariety (fromKey keyVariety k)) ++ " backend"
 		)
 	
-	retrieveKeyFileFromImport dbv ciddbv k af dest p =
-		getkeycids ciddbv k >>= \case
-			(cid:_) -> do
-				l <- getfirstexportloc dbv k
-				snd <$> retrieveExportWithContentIdentifier (importActions r) l cid dest (Left k) p
+	retrieveFromImport getlocs ciddbv k af dest p = do
+		cids <- getkeycids ciddbv k
+		if not (null cids)
+			then getlocs $ \loc ->
+				snd <$> retrieveExportWithContentIdentifier (importActions r) loc cids dest (Left k) p
 			-- In case a content identifier is somehow missing,
 			-- try this instead.
-			[] -> if isexport
-				then retrieveKeyFileFromExport dbv k af dest p
+			else if isexport
+				then retrieveFromExport getlocs k af dest p
 				else giveup "no content identifier is recorded, unable to retrieve"
-	
-	-- versionedExport remotes have a key/value store, so can use
-	-- the usual retrieveKeyFile, rather than an import/export
-	-- variant. However, fall back to that if retrieveKeyFile fails.
-	supportversionedretrieve k af dest p vc a
-		| versionedExport (exportActions r) =
-			retrieveKeyFile r k af dest p vc
-				`catchNonAsync` const a
-		| otherwise = a
+
+	checkpresentwith k a = ifM a
+		( return True
+		, if annexobjects
+			then checkpresentannexobject k
+			else return False
+		)
 
 	checkPresentImport ciddbv k loc =
 		checkPresentExportWithContentIdentifier
 			(importActions r)
 			k loc 
 			=<< getkeycids ciddbv k
+
+	annexobjectlocation k = exportAnnexObjectLocation gc k
+
+	checkpresentannexobject k = 
+		checkPresentExport (exportActions r) k (annexobjectlocation k)
+
+	storeannexobject k o p = prepSendAnnex' k o >>= \case
+		Nothing -> giveup "content is not available"
+		Just (src, _, checkmodified) -> do
+			let loc = annexobjectlocation k
+			storeExport (exportActions r) src k loc p
+			checkmodified >>= \case
+				Nothing -> return ()
+				Just err -> do
+					removeExport (exportActions r) k loc
+					giveup err
+
+	removeannexobject dbv k = 
+		getanyexportlocs dbv k >>= \case
+			[] -> ifM (isexportinconflict dbv)
+				( do
+					warnExportImportConflict r
+					giveup "Cannot remove content from the remote until the conflict has been resolved."
+				, removeExport (exportActions r) k (annexobjectlocation k)
+				)
+			_ -> giveup "This key is part of the exported tree, so can only be removed by exporting a tree that does not include it."
+
+	retrieveannexobject k af dest p =
+		retrieveFromExport getlocs k af dest p
+	  where
+		getlocs a = a (annexobjectlocation k)
+
+	supportretrieveannexobject dbv k af dest p a
+		| annexobjects = tryNonAsync a >>= \case
+			Right res -> return res
+			Left err -> tryNonAsync (retrieveannexobject k af dest p) >>= \case
+				Right res -> return res
+				-- Both failed, so which exception to
+				-- throw? If there are known export
+				-- locations, throw the exception from
+				-- retrieving from the export locations.
+				-- If there are no known export locations,
+				-- throw the exception from retrieving from
+				-- the annexobjects location.
+				Left err' -> getanyexportlocs dbv k >>= \case
+					[] -> ifM (isexportinconflict dbv)
+						( throwM err
+						, throwM err'
+						)
+					_ -> throwM err
+		| otherwise = a
+
+	-- versioned remotes have a key/value store which
+	-- the usual retrieveKeyFile can be used with, rather than
+	-- an import/export variant. However, fall back to that
+	-- if retrieveKeyFile fails.
+	supportversionedretrieve k af dest p vc a
+		| versioned =
+			retrieveKeyFile r k af dest p vc
+				`catchNonAsync` const a
+		| otherwise = a
