@@ -1,6 +1,6 @@
 {- HttpAlso remote (readonly).
  -
- - Copyright 2020-2023 Joey Hess <id@joeyh.name>
+ - Copyright 2020-2025 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
@@ -24,6 +24,7 @@ import Utility.Metered
 import Annex.Verify
 import qualified Annex.Url as Url
 import Annex.SpecialRemote.Config
+import Git.FilePath
 
 import Data.Either
 import qualified Data.Map as M
@@ -56,9 +57,9 @@ gen r u rc gc rs = do
 	ll <- liftIO newLearnedLayout
 	return $ Just $ specialRemote c
 		cannotModify
-		(downloadKey url ll)
+		(downloadKey gc url ll)
 		cannotModify
-		(checkKey url ll)
+		(checkKey gc url ll)
 		(this url c cst)
   where
 	this url c cst = Remote
@@ -67,6 +68,7 @@ gen r u rc gc rs = do
 		, name = Git.repoDescribe r
 		, storeKey = cannotModify
 		, retrieveKeyFile = retrieveKeyFileDummy
+		, retrieveKeyFileInOrder = pure True
 		, retrieveKeyFileCheap = Nothing
 		-- HttpManagerRestricted is used here, so this is
 		-- secure.
@@ -77,10 +79,9 @@ gen r u rc gc rs = do
 		, checkPresentCheap = False
 		, exportActions = ExportActions
 			{ storeExport = cannotModify
-			, retrieveExport = retriveExportHttpAlso url
+			, retrieveExport = retriveExportHttpAlso gc url
 			, removeExport = cannotModify
-			, versionedExport = False
-			, checkPresentExport = checkPresentExportHttpAlso url
+			, checkPresentExport = checkPresentExportHttpAlso gc url
 			, removeExportDirectory = Nothing
 			, renameExport = cannotModify
 			}
@@ -95,7 +96,7 @@ gen r u rc gc rs = do
 		, readonly = True
 		, appendonly = False
 		, untrustworthy = False
-		, availability = GloballyAvailable
+		, availability = pure GloballyAvailable
 		, remotetype = remote
 		, mkUnavailable = return Nothing
 		, getInfo = return []
@@ -107,47 +108,48 @@ gen r u rc gc rs = do
 cannotModify :: a
 cannotModify = giveup "httpalso special remote is read only"
 
-httpAlsoSetup :: SetupStage -> Maybe UUID -> Maybe CredPair -> RemoteConfig -> RemoteGitConfig -> Annex (RemoteConfig, UUID)
-httpAlsoSetup _ Nothing _ _ _ =
+httpAlsoSetup :: SetupStage -> Maybe UUID -> RemoteName -> Maybe CredPair -> RemoteConfig -> RemoteGitConfig -> Annex (RemoteConfig, UUID)
+httpAlsoSetup _ Nothing _ _ _ _ =
 	giveup "Must use --sameas when initializing a httpalso remote."
-httpAlsoSetup _ (Just u) _ c gc = do
+httpAlsoSetup ss (Just u) _ _ c gc = do
 	_url <- maybe (giveup "Specify url=")
 		(return . fromProposedAccepted)
 		(M.lookup urlField c)
 	c' <- if isJust (M.lookup encryptionField c)
-		then fst <$> encryptionSetup c gc
+		then fst <$> encryptionSetup ss c gc
 		else pure c
 	gitConfigSpecialRemote u c' [("httpalso", "true")]
 	return (c', u)
 
-downloadKey :: Maybe URLString -> LearnedLayout -> Retriever
-downloadKey baseurl ll = fileRetriever' $ \dest key p iv ->
-	downloadAction (fromRawFilePath dest) p iv (keyUrlAction baseurl ll key)
+downloadKey :: RemoteGitConfig -> Maybe URLString -> LearnedLayout -> Retriever
+downloadKey gc baseurl ll = fileRetriever' $ \dest key p iv ->
+	downloadAction gc dest p iv (keyUrlAction baseurl ll key)
 
-retriveExportHttpAlso :: Maybe URLString -> Key -> ExportLocation -> FilePath -> MeterUpdate -> Annex Verification
-retriveExportHttpAlso baseurl key loc dest p = do
+retriveExportHttpAlso :: RemoteGitConfig -> Maybe URLString -> Key -> ExportLocation -> OsPath -> MeterUpdate -> Annex Verification
+retriveExportHttpAlso gc baseurl key loc dest p = do
 	verifyKeyContentIncrementally AlwaysVerify key $ \iv ->
-		downloadAction dest p iv (exportLocationUrlAction baseurl loc)
+		downloadAction gc dest p iv (exportLocationUrlAction baseurl loc)
 
-downloadAction :: FilePath -> MeterUpdate -> Maybe IncrementalVerifier -> ((URLString -> Annex (Either String ())) -> Annex (Either String ())) -> Annex ()
-downloadAction dest p iv run =
-	Url.withUrlOptions $ \uo ->
+downloadAction :: RemoteGitConfig -> OsPath -> MeterUpdate -> Maybe IncrementalVerifier -> ((URLString -> Annex (Either String ())) -> Annex (Either String ())) -> Annex ()
+downloadAction gc dest p iv run =
+	Url.withUrlOptions (Just gc) $ \uo ->
 		run (\url -> Url.download' p iv url dest uo)
 			>>= either giveup (const (return ()))
 
-checkKey :: Maybe URLString -> LearnedLayout -> CheckPresent
-checkKey baseurl ll key =
-	isRight <$> keyUrlAction baseurl ll key (checkKey' key)
+checkKey :: RemoteGitConfig -> Maybe URLString -> LearnedLayout -> CheckPresent
+checkKey gc baseurl ll key =
+	isRight <$> keyUrlAction baseurl ll key (checkKey' gc key)
 
-checkKey' :: Key -> URLString -> Annex (Either String ())
-checkKey' key url = ifM (Url.withUrlOptions $ Url.checkBoth url (fromKey keySize key))
-	( return (Right ())
-	, return (Left "content not found")
-	)
+checkKey' :: RemoteGitConfig -> Key -> URLString -> Annex (Either String ())
+checkKey' gc key url = 
+	ifM (Url.withUrlOptions (Just gc) $ Url.checkBoth url (fromKey keySize key))
+		( return (Right ())
+		, return (Left "content not found")
+		)
 
-checkPresentExportHttpAlso :: Maybe URLString -> Key -> ExportLocation -> Annex Bool
-checkPresentExportHttpAlso baseurl key loc =
-	isRight <$> exportLocationUrlAction baseurl loc (checkKey' key)
+checkPresentExportHttpAlso :: RemoteGitConfig -> Maybe URLString -> Key -> ExportLocation -> Annex Bool
+checkPresentExportHttpAlso gc baseurl key loc =
+	isRight <$> exportLocationUrlAction baseurl loc (checkKey' gc key)
 
 type LearnedLayout = TVar (Maybe [Key -> URLString])
 
@@ -192,7 +194,7 @@ exportLocationUrlAction
 	-> (URLString -> Annex (Either String ()))
 	-> Annex (Either String ())
 exportLocationUrlAction (Just baseurl) loc a =
-	a (baseurl P.</> fromRawFilePath (fromExportLocation loc))
+	a (baseurl P.</> fromOsPath (fromExportLocation loc))
 exportLocationUrlAction Nothing _ _ = noBaseUrlError
 
 -- cannot normally happen
@@ -228,5 +230,10 @@ supportedLayouts baseurl =
 	  ]
 	]
   where
-	mkurl k hasher = baseurl P.</> fromRawFilePath (hasher k) P.</> kf k
-	kf k = fromRawFilePath (keyFile k)
+	mkurl k hasher = baseurl
+		-- On windows, the hasher uses `\` path separators,
+		-- but for an url, it needs to use '/'.
+		-- So, use toInternalGitPath.
+		P.</> fromOsPath (toInternalGitPath (hasher k)) 
+		P.</> kf k
+	kf k = fromOsPath (keyFile k)
